// ===================================================================================
// Copyright © Genetec Inc., 2020
// All rights reserved.
// May only be used in accordance with a valid Source Code License Agreement.
//
// Genetec reserves the right to refuse to provide technical support in the event the 
// macro has been modified in any way, and for any reason.
// ===================================================================================
// ===================================================================================
// Contagion/Contaminant Proximity Report
//      Version 1.5
// ===================================================================================

using Genetec.Sdk;
using Genetec.Sdk.Entities;
using Genetec.Sdk.Queries;
using Genetec.Sdk.Scripting;
using System;
using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Genetec.Sdk.Entities.Utilities;
public sealed class Contagion : UserMacro
{
    public Guid Cardholder { get; set; }
    public DateTime StartTime { get; set; }
    public DateTime EndTime { get; set; }
    public int MinutesWindow { get; set; }
    public Guid EmailRecipient { get; set; }
    public bool SaveToDisk { get; set; }
    private string reportName;
    private HashSet<Tuple<Guid, DateTime>> m_contagionLocationTimes = new HashSet<Tuple<Guid, DateTime>>();
    private StringBuilder m_results = new StringBuilder();
    private ManualResetEvent m_prefetchFinishedSignalEvent = new ManualResetEvent(false);
    private ManualResetEvent m_contagionQueryFinishedSignalEvent = new ManualResetEvent(false);
    public override void Execute()
    {
        try
        {
            if (EmailRecipient != Guid.Empty && !IsEmailValid())
            {
                Sdk.ActionManager.SendMessage(this.InstigatorGuid, "Invalid email address.");
            }

            var cardholderEntity = Sdk.GetEntity(Cardholder) as Cardholder;
            var cardholderName = cardholderEntity.FirstName + " " + cardholderEntity.LastName;
            reportName = "Synergis Proximity Report " + cardholderName + ".csv";
            // Prefetch cardholders and doors to avoid scalability breaches
            m_results.AppendLine(String.Format("\"Cardholder:\",\"{0}\"", cardholderName));
            m_results.AppendLine(String.Format("\"Dates searched:\",\"{0} - {1}\"", StartTime.ToString("yyyy/MM/dd HH:mm:ss"), EndTime.ToString("yyyy/MM/dd HH:mm:ss")));
            m_results.AppendLine(String.Format("\"Time window:\",\"{0} minutes\"", MinutesWindow));
            m_results.AppendLine("\r\n");
            GetAllEntities(new List<EntityType> { EntityType.Door, EntityType.Cardholder });
            m_prefetchFinishedSignalEvent.WaitOne();
            var cardholderActivityReport = Sdk.ReportManager.CreateReportQuery(Genetec.Sdk.ReportType.CardholderActivity) as CardholderActivityQuery;
            cardholderActivityReport.Events.Add(EventType.CardholderAccessGranted);
            cardholderActivityReport.QueryEntities.Add(Cardholder);
            cardholderActivityReport.TimeRange.SetTimeRange(StartTime, EndTime);
            cardholderActivityReport.BeginQuery(OnInitialQueryCompleted, OnInitialQueryResultsReceived, cardholderActivityReport);
            m_contagionQueryFinishedSignalEvent.WaitOne();
            foreach (var contagionPoint in m_contagionLocationTimes)
            {
                var entity = Sdk.GetEntity(contagionPoint.Item1);
                if (entity == null)
                {
                    m_results.AppendLine("Error, entity null " + contagionPoint.Item1.ToString());
                    continue;
                }
                var name = entity.Name;
                var header = String.Format("\"{0}\",\"{1}\"", name, contagionPoint.Item2.ToString("yyyy/MM/dd HH:mm:ss"));
                m_results.AppendLine(header);
                m_results.AppendLine("");
                var doorActivityReport = Sdk.ReportManager.CreateReportQuery(Genetec.Sdk.ReportType.DoorActivity) as DoorActivityQuery;
                doorActivityReport.Events.Add(EventType.CardholderAccessGranted);
                doorActivityReport.QueryEntities.Add(contagionPoint.Item1);
                doorActivityReport.TimeRange.SetTimeRange(contagionPoint.Item2.AddMinutes(0 - MinutesWindow), contagionPoint.Item2.AddMinutes(MinutesWindow));
                var results = doorActivityReport.Query();
                var resultSet = new HashSet<string>();
                foreach (DataRow dr in results.Data.Rows)
                {
                    if ((EventType)dr[1] != EventType.AccessGranted) continue;
                    var chGuid = (Guid)dr[7];
                    if (chGuid.Equals(Cardholder) || chGuid.Equals(Guid.Empty)) continue;
                    var date = ((DateTime)dr[0]);
                    var localDate = TimeZoneInfo.ConvertTime(date, TimeZoneInfo.FindSystemTimeZoneById((string)dr[9]));
                    var chEntity = Sdk.GetEntity(chGuid) as Cardholder;
                    if (chEntity == null) continue;
                    var chName = chEntity.FirstName + " " + chEntity.LastName;
                    var differential = localDate.Subtract(contagionPoint.Item2);
                    var differentialString = differential.Ticks > 0
                        ? "+" + differential.ToString(@"hh\:mm\:ss")
                        : "-" + differential.ToString(@"hh\:mm\:ss");
                    resultSet.Add(String.Format("\"{0}\",\"{1}\",\"{2}\"", chName, localDate.ToString("yyyy/MM/dd HH:mm:ss"), differentialString));
                }
                foreach (var line in resultSet)
                {
                    m_results.AppendLine(line);
                }
                m_results.AppendLine("\r\n\r\n");
            }

            if (EmailRecipient != Guid.Empty && IsEmailValid())
            {
                var recipient = Sdk.GetEntity(EmailRecipient) as Entity;
                var content = Encoding.UTF8.GetBytes(m_results.ToString());
                var attachment = new EmailAttachment(reportName, content);
                Sdk.ActionManager.SendEmail(EmailRecipient, reportName, attachment);

                if (SaveToDisk)
                {
                    System.IO.File.WriteAllText(reportName, m_results.ToString());
                    var message = string.Format("{0} emailed to {1} and saved to disk.", reportName, recipient.Name);
                    Sdk.ActionManager.SendMessage(this.InstigatorGuid, message);
                }
                else
                {
                    var message = string.Format("{0} emailed to {1}.", reportName, recipient.Name);
                    Sdk.ActionManager.SendMessage(this.InstigatorGuid, message);
                }

            }
            else if (SaveToDisk)
            {
                System.IO.File.WriteAllText(reportName, m_results.ToString());
                var message = string.Format("{0} saved to disk.", reportName);
                Sdk.ActionManager.SendMessage(this.InstigatorGuid, message);
            }
        }
        catch (Exception ex)
        {
            Sdk.ActionManager.SendMessage(this.InstigatorGuid, ex.ToString());
        }


    }
    private void OnInitialQueryResultsReceived(QueryCompletedEventArgs obj)
    {
    }
    public void GetAllEntities(List<EntityType> types, int pageNumber = 1, int pageSize = 500)
    {
        var entityConfigurationQuery = Sdk.ReportManager.CreateReportQuery(ReportType.EntityConfiguration) as EntityConfigurationQuery;
        foreach (EntityType entityType in types)
        {
            entityConfigurationQuery.EntityTypeFilter.Add(entityType);
        }
        entityConfigurationQuery.Page = pageNumber;
        entityConfigurationQuery.PageSize = pageSize;
        entityConfigurationQuery.BeginQuery(OnGetAllEntitiesQueryFinished, entityConfigurationQuery);
    }
    public void OnGetAllEntitiesQueryFinished(IAsyncResult ar)
    {
        var query = ar.AsyncState as EntityConfigurationQuery;
        var results = query.EndQuery(ar);
        if (results.Data.Rows.Count > query.PageSize)
        {
            query.Page++;
            GetAllEntities(query.EntityTypeFilter.ToList(), query.Page);
        }
        else
        {
            m_prefetchFinishedSignalEvent.Set();
        }
    }
    private void OnInitialQueryCompleted(IAsyncResult ar)
    {
        var query = ar.AsyncState as CardholderActivityQuery;
        if (query != null)
        {
            var allResults = query.EndQuery(ar);
            foreach (DataRow dr in allResults.Data.Rows)
            {
                if ((EventType)dr[1] == EventType.AccessGranted || (EventType)dr[1] == EventType.CardholderAccessGranted)
                {
                    var apGuid = (Guid)dr[11];
                    var entityType = Sdk.GetEntity(apGuid).EntityType;
                    var date = ((DateTime)dr[0]);
                    var localDate = TimeZoneInfo.ConvertTime(date, TimeZoneInfo.FindSystemTimeZoneById((string)dr[9]));
                    if (entityType == EntityType.Door || entityType == EntityType.Elevator)
                        m_contagionLocationTimes.Add(Tuple.Create(apGuid, localDate));
                }
            }
            m_contagionQueryFinishedSignalEvent.Set();
        }
    }

    public bool IsEmailValid()
    {
        bool valid = false;
        var ent = Sdk.GetEntity(EmailRecipient) as Entity;
        switch (ent.EntityType)
        {
            case EntityType.User:
                var user = Sdk.GetEntity(EmailRecipient) as User;
                if (!string.IsNullOrEmpty(user.EmailAddress))
                    valid = true;
                break;

            case EntityType.UserGroup:
                var usergr = Sdk.GetEntity(EmailRecipient) as UserGroup;
                if (!string.IsNullOrEmpty(usergr.EmailAddress))
                    valid = true;
                break;

            case EntityType.Cardholder:
                var cr = Sdk.GetEntity(EmailRecipient) as Cardholder;
                if (!string.IsNullOrEmpty(cr.EmailAddress))
                    valid = true;
                break;

            case EntityType.CardholderGroup:
                var gr = Sdk.GetEntity(EmailRecipient) as CardholderGroup;
                if (!string.IsNullOrEmpty(gr.EmailAddress))
                    valid = true;
                break;
            default:
                break;
        }
        return valid;
    }
    protected override void CleanUp()
    {
    }
}
